-- ToME - Tales of Maj'Eyal
-- Copyright (C) 2009, 2010, 2011, 2012, 2013 Nicolas Casalini
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org


-- damage: initial physical damage and used for fractional knockback damage
-- knockback: distance to knockback
-- knockbackDamage: when knockback strikes something, both parties take damage - percent of damage * remaining knockback
-- power: used to determine the initial radius of particles
local function forceHit(self, t, target, sourceX, sourceY, damage, knockback, knockbackDamage, power, max)
	-- apply initial damage
	if damage > 0 then
		damage = self:mindCrit(damage)
		self:project({type="hit", range=10, talent=t}, target.x, target.y, DamageType.NATURE, damage)
		game.level.map:particleEmitter(target.x, target.y, 1, "force_hit", {power=power, dx=target.x - sourceX, dy=target.y - sourceY})
	end

	-- knockback?
	if not target.dead and knockback and knockback > 0 and target:canBe("knockback") and (target.never_move or 0) < 1 then
		-- give direct hit a direction?
		if sourceX == target.x and sourceY == target.y then
			local newDirection = rng.table(util.adjacentDirs())
			local dx, dy = util.dirToCoord(newDirection, sourceX, sourceY)
			sourceX = sourceX + dx
			sourceY = sourceY + dy
		end

		local block_actor = function(_, bx, by) return game.level.map:checkEntity(bx, by, Map.TERRAIN, "block_move", target) end
		local lineFunction = core.fov.line(sourceX, sourceY, target.x, target.y, block_actor, true)
		local finalX, finalY = target.x, target.y
		local knockbackCount = 0
		local blocked = false
		while knockback > 0 do
			local x, y, is_corner_blocked = lineFunction:step(true)

			if not game.level.map:isBound(x, y) or is_corner_blocked or game.level.map:checkAllEntities(x, y, "block_move", target) then
				-- blocked
				local nextTarget = game.level.map(x, y, Map.ACTOR)
				if nextTarget then
					if knockbackCount > 0 then

					target:logCombat(nextTarget, "#Source# was flung %d spaces into #Target#!", knockbackCount)
					else
						target:logCombat(nextTarget, "#Source# was flung into #Target#!")
					end
				elseif knockbackCount > 0 then
					game.logSeen(target, "%s was flung back %d spaces!", target.name:capitalize(), knockbackCount)
				else
					game.logSeen(target, "%s was flung!", target.name:capitalize())
				end

				-- take partial damage
				local blockDamage = damage * util.bound(knockback * (knockbackDamage / 100), 0, 1.5)
				self:project({type="hit", range=10, talent=t}, target.x, target.y, DamageType.PHYSICAL, blockDamage)

				if nextTarget then
					-- start a new force hit with the knockback damage and current knockback
					if max > 0 then
						forceHit(self, t, nextTarget, sourceX, sourceY, blockDamage, knockback, knockbackDamage, power / 2, max - 1)
					end
				end

				knockback = 0
				blocked = true
			else
				-- allow move
				finalX, finalY = x, y
				knockback = knockback - 1
				knockbackCount = knockbackCount + 1
			end
		end

		if not blocked and knockbackCount > 0 then
			game.logSeen(target, "%s was flung back %d spaces!", target.name:capitalize(), knockbackCount)
		end

		if not target.dead and (finalX ~= target.x or finalY ~= target.y) then
			local ox, oy = target.x, target.y
			target:move(finalX, finalY, true)
			if config.settings.tome.smooth_move > 0 then
				target:resetMoveAnim()
				target:setMoveAnim(ox, oy, 9, 5)
			end
		end
	end
end



newTalent{
	short_name = "HIDDEN_PATH",
	name = "Hidden Path",
	type = {"wild-gift/roots", 1},
	require = gifts_req1,
	random_ego = "escape",
	points = 5,
	cooldown = 5,
	equilibrium = 5,	
	tactical = { CLOSEIN = 2, ESCAPE = 2 },
	range = function(self, t) return math.floor(self:combatTalentScale(t, 4, 8)) end,
	requires_target = true,
	target = function(self, t)
		return {type="hit", range=self:getTalentRange(t), nolock=true, nowarning=true}
	end,
	direct_hit = true,
	no_energy = false,
	is_teleport = false,
	action = function(self, t)
		local tg = self:getTalentTarget(t)
		local x, y = self:getTarget(tg)
		if not x or not y then return nil end
		if not self:hasLOS(x, y) or game.level.map:checkEntity(x, y, Map.TERRAIN, "block_move") then
			game.logSeen(self, "You do not have line of sight.")
			return nil
		end
		local __, x, y = self:canProject(tg, x, y)

		game.level.map:particleEmitter(self.x, self.y, 1, "slime")

		-- since we're using a precise teleport we'll look for a free grid first
		local tx, ty = util.findFreeGrid(x, y, 5, true, {[Map.ACTOR]=true})
		if tx and ty then
			if not self:teleportRandom(tx, ty, 0) then
				game.logSeen(self, "The roots crumble!")
			end
		end

		game.level.map:particleEmitter(self.x, self.y, 1, "slime")
		game:playSoundNear(self, "talents/teleport")

		return true
	end,
	info = function(self, t)
		local range = self:getTalentRange(t)
		return ([[Roots drag you underneath the earth to target location up to %d tiles away within line of sight.  Additional talent points increase the range.]]):format(range)
	end,
}

newTalent{
	short_name = "FLING",
	name = "Fling",
	type = {"wild-gift/roots", 2},
	require = gifts_req2,
	random_ego = "attack",
	points = 5,
	cooldown = 5,
	equilibrium = 5,
	tactical = {ATTACK = { PHYSICAL = 2 } } ,
	range = 5,
	requires_target = true,
	direct_hit = true,
	getDamage = function(self, t)
		return self:combatTalentMindDamage(t, 0, 140)
	end,
	getKnockback = function(self, t)
		return self:combatTalentScale(t, 2, 5)
	end,
	action = function(self, t)
		local range = self:getTalentRange(t)

		local tg = {type="hit", range=self:getTalentRange(t)}
		local x, y, target = self:getTarget(tg)
		if not x or not y or not target or core.fov.distance(self.x, self.y, x, y) > range then return nil end

		--local distance = math.max(1, core.fov.distance(self.x, self.y, x, y))
		local power = 1 --(1 - ((distance - 1) / range))
		local damage = t.getDamage(self, t) * power
		local knockback = t.getKnockback(self, t)
		forceHit(self, t, target, self.x, self.y, damage, knockback, 7, power, 10)
		return true
	end,
	
	info = function(self, t)
		local damage = t.getDamage(self, t)
		local knockback = t.getKnockback(self, t)
		return ([[Roots grab the target for %d damage and fling it back %d spaces.]]):format(self.damDesc(self, DamageType.NATURE, damage), knockback)
	end,
}

newTalent{
	short_name = "SINKHOLE",
	name = "Sink Hole",
	type = {"wild-gift/roots", 3},
	require = gifts_req3,
	random_ego = "defensive",
	points = 5,
	cooldown = 10,
	equilibrium = 10,
	tactical = { ATTACKAREA = {NATURE = 2}, DISABLE = { pin = 2 } },
	range = 8,
	radius = function(self, t) return math.floor(self:combatTalentScale(t, 2.3, 3.7)) end,
	direct_hit = true,
	requires_target = true,
	target = function(self, t)
		return {type="ball", range=self:getTalentRange(t), radius=self:getTalentRadius(t), friendlyfire = false}
	end,
	getDamage = function(self, t)
		return self:combatTalentMindDamage(t, 0, 280)
	end,	
	getDur = function(self,t)
		return self:combatTalentScale(t, 5, 8)
		
	end,
	action = function(self, t)
		local tg = self:getTalentTarget(t)
		local x, y = self:getTarget(tg)
		if not x or not y then return nil end
		local _ _, _, _, x, y = self:canProject(tg, x, y)
		local grids = self:project(tg, x, y, function(px, py)
			local target = game.level.map(px, py, Map.ACTOR)
			if not target then return end
			local tx, ty = util.findFreeGrid(x, y, 5, true, {[Map.ACTOR]=true})
			if target:canBe("pin") then
				target:setEffect(target.EFF_PINNED, t.getDur(self, t), {apply_power=t.getDamage(self, t)})
			else
				game.logSeen(target, "%s resists!", target.name:capitalize())
			end
		end)
		self:project (tg, x, y, DamageType.NATURE, self:mindCrit(t.getDamage(self, t)))

		game:playSoundNear(self, "talents/earth")
		return true
	end,	
	info = function(self, t)
		return ([[Roots erupt from the earth to attack the targeted area for %d and attempting to pin foes within radius %d .
		]]):format(self.damDesc(self, DamageType.NATURE, t.getDamage(self, t)), self:getTalentRadius(t))
	end,
}

newTalent{
	short_name = "SPROUT",
	name = "Sprout",
	type = {"wild-gift/roots", 4},
	require = gifts_req4,
	random_ego = "defensive",
	points = 5,
	cooldown = 20,
	equilibrium = 10,
	range = 10,
	requires_target = true,
	tactical = { DISABLE = 2 },
	getDuration = function(self, t) return math.floor(self:combatTalentScale(t, 5, 9)) end,
	getLength = function(self, t) return 1 + math.floor(self:combatTalentScale(t, 3, 7)/2)*2 end,
	action = function(self, t)
		local halflength = math.floor(t.getLength(self,t)/2)
		local block = function(_, lx, ly)
			return game.level.map:checkAllEntities(lx, ly, "block_move")
		end
		local tg = {type="wall", range=self:getTalentRange(t), halflength=halflength, talent=t, halfmax_spots=halflength+1, block_radius=block}
		local x, y = self:getTarget(tg)
		if not x or not y then return nil end
		local _ _, _, _, x, y = self:canProject(tg, x, y)
		if game.level.map:checkEntity(x, y, Map.TERRAIN, "block_move") then return nil end

		self:project(tg, x, y, function(px, py, tg, self)
			local oe = game.level.map(px, py, Map.TERRAIN)
			if not oe or oe.special then return end
			if not oe or oe:attr("temporary") or game.level.map:checkAllEntities(px, py, "block_move") then return end
			local Object = require "mod.class.Object"
			local e = Object.new{
				old_feat = oe,
				name = "Root Wall", image = "npc/roots.png",
				desc = "a summoned, wall of roots",
				type = "wall",
				display = '#', color=colors.LIGHT_BLUE, back_color=colors.BLUE,
				always_remember = true,
				can_pass = {pass_wall=1},
				does_block_move = true,
				show_tooltip = true,
				block_move = true,
				block_sight = false,
				temporary = 4 + self:getTalentLevel(t),
				x =px, y = py,
				canAct = false,
				act = function(self)
					self:useEnergy()
					self.temporary = self.temporary - 1
					if self.temporary <= 0 then
						game.level.map(self.x, self.y, engine.Map.TERRAIN, self.old_feat)
						game.level:removeEntity(self)
						game.level.map:updateMap(self.x, self.y)
					end
				end,
				dig = function(src, x, y, old)
					game.level:removeEntity(old)
					return nil, old.old_feat
				end,
				--summoner_gain_exp = true,
				--summoner = self,
			}
			e.tooltip = mod.class.Grid.tooltip
			game.level:addEntity(e)
			game.level.map(px, py, Map.TERRAIN, e)
		end)
		return true
	end,
	info = function(self, t)
		return ([[Summons a wall of roots of %d length for %d turns. The wall is transparent.
		]]):format(3 + math.floor(self:getTalentLevel(t) / 2) * 2, t.getDuration(self, t))
	end,
}


